open Printf
open Ext

(* Globals used for command line parameters *)
let acfile = ref ""
let qfile = ref ""
let evfile = ref ""
let marg_outfile = ref ""
let sameev = ref false
let marg = ref false
let marg_outfile = ref ""
let preprune = ref false

let load_circuit filename =
  try Circuit.load (open_in filename) 
  with Circuit.Unknown_child(n,c) -> 
         nlogf "Error: node %s references unknown child node %s.\n" n c; 
         exit 0 

let usage = "Usage: acquery -c <ac file> -q <query> -ev <evidence> [...]"
let args = Arg.align
  ([("-c", Arg.Set_string acfile, " Input arithmetic circuit");
    ("-ev", Arg.Set_string evfile, " Evidence file");
    ("-q",  Arg.Set_string qfile,  " Query file");
    ("-sameev", Arg.Set sameev, " Use the same evidence for all queries");
    ("-preprune", Arg.Set preprune, " Prune circuit for the evidence");
    ("-m", Arg.Set marg, " Use conditional marginals instead of joint distribution");
    ("-mo", Arg.Set_string marg_outfile, " Print marginals to specified file")]
    @ common_arguments)

let answer_joint_queries circ query_buf evidence_buf =
  let scratch = Circuit.create_scratch circ in

  (* Handle shared evidence: 
     Read just one piece of evidence and compute its probability only once. *)
  if !evfile = "" then sameev := true;
  let sharedev =
    if !sameev && !evfile <> "" then 
      Data.input_example evidence_buf
    else
      [||] in
  let logp_shared_ev = 
    if !sameev && !evfile <> "" then begin
      (* Pruning the circuit for the evidence may speed up resulting
       * queries significantly, by shrinking the size of the circuit. *)
      if !preprune then
        Circuit.prune_for_evidence circ sharedev; 
      Circuit.logprob_x scratch circ sharedev 
    end else
      0.0 in

  (* Compute each log conditional probability... *)
  let sl = stats_make () in
  let st = stats_make () in
  begin try 
    while true do 
      (* By definition of conditional probability:
           log P(q|e) = log P(q) - log P(e) *)
      start_time "query";
      let q = Data.input_example query_buf in
      (* Input validation *)
      Data.check_point circ.Circuit.schema q;
      let e = if !sameev then sharedev 
              else Data.input_example evidence_buf in

      (* Check for compatibility between evidence and query. *)
      if Array.length e > 0 then
        Data.check_point circ.Circuit.schema e;
      let mismatch = ref false in
      for i = 0 to Array.length e - 1 do
        if e.(i) >= 0 then begin
          if q.(i) < 0 then 
            q.(i) <- e.(i)
          else if q.(i) != e.(i) then
            mismatch := true
        end
      done;

      (* Compute probability of query and evidence *)
      let logp_q = if !mismatch then log 0. 
                   else Circuit.logprob_x scratch circ q in
      let logp_ev = if !sameev then logp_shared_ev
                    else Circuit.logprob_x scratch circ e in
      let l = logp_q -. logp_ev in
      let dt = elapsed "query" in
      if log_exists log_verbose then
        vlogf "%f %f\n" l dt
      else
        nlogf "%f\n" l;
      stats_add sl l;
      stats_add st dt
    done 
  with Data.Eof -> () end;
  (sl, st)


let answer_marg_queries circ query_buf evidence_buf =
  let ds = Circuit.create_deriv_scratch circ in

  (* Handle shared evidence *)
  if !evfile = "" then sameev := true;
  let sharedlogmarg =
    if !sameev then begin
      start_time "marg";
      let e = if !evfile = "" then [||] 
              else Data.input_example evidence_buf in
      let m = Circuit.get_logmarginals circ ds e in
      nlogf "AC inference time: %fs\n" (elapsed "marg");
      m
    end else
      [||] in
  let mout =
    if !marg_outfile <> "" then open_out !marg_outfile 
    else log_stream log_normal in
  let sl = stats_make () in
  let st = stats_make () in
  begin try 
    while true do
      start_time "query";
      let logmarg = 
        if !sameev then
          sharedlogmarg 
        else
          let e = Data.input_example evidence_buf in
          Circuit.get_logmarginals circ ds e in

      if !qfile = "" || !marg_outfile <> "" then begin
        let rawmarg = Array.map (Array.map exp) logmarg in
        Data.output_marginals mout rawmarg
      end;

      if !qfile = "" then raise Data.Eof;

      let q = Data.input_example query_buf in
      let llvar var vmarg =
        if q.(var) >= 0 then vmarg.(q.(var)) else log_one in
      let l = Array.sumf (Array.mapi llvar logmarg) in
      let dt = elapsed "query" in
      if log_exists log_verbose then
        vlogf "%f %f\n" l dt
      else
        nlogf "%f\n" l;
      stats_add sl l;
      stats_add st dt
    done
  with Data.Eof -> () end;
  (sl, st)


let main () = 
  start_time "total";

  (* Parse arguments *)
  Arg.parse args ignore usage;
  if !marg_outfile <> "" then marg := true;
  if !acfile = "" || (!qfile = "" && !marg = false) then
    (Arg.usage args usage; exit 0);

  common_log_init ();

  let circ = load_circuit !acfile in
  Circuit.update_parents circ;
  let evstream = if !evfile <> "" then open_in !evfile else stdin in 
  let qstream = if !qfile <> "" then open_in !qfile else stdin in 

  (* Run inference *)
  let (sl, st) =
    if !marg then
      answer_marg_queries circ qstream evstream 
    else
      answer_joint_queries circ qstream evstream in

  (* Print stats *)
  if stats_n sl > 0 then begin
    if classify_float (stats_stderr st) = FP_nan then
      vlogf "time = %f +/- 0\n" (stats_mean st)
    else
      vlogf "time = %f +/- %f\n" (stats_mean st) (stats_stderr st);
    nlogf "avg = %f +/- %f\n" (stats_mean sl) (stats_stderr sl);
  end;
  vlogf "Total time: %fs\n" (elapsed "total")

;;
main ()
