open Printf
open Ext

(* Globals used for command line parameters *)
let modelfile = ref ""
let outfile = ref ""
let evfile = ref ""
let tofeatures = ref false

let usage = "Usage: mconvert -m <model> -o <output>"
let args = Arg.align
 ([("-m", Arg.Set_string modelfile, " Input model file (MN, BIF, XMOD, or AC)");
   ("-o", Arg.Set_string outfile, " Output model file");
   ("-ev", Arg.Set_string evfile, " Evidence file");
   ("-feat", Arg.Set tofeatures, " Convert MN factors to features")]
   @ common_arguments)

let main () = 
  Arg.parse args ignore usage;
  if !modelfile = "" || !outfile = "" then
    (Arg.usage args usage; exit 0);

  common_log_init();

  let has_ev = !evfile <> "" in
  let ev = if has_ev then Data.input_example (open_in !evfile)
           else [||] in

  let proc_mn mn =
    let mn = if has_ev then Mn.simplify mn ev else mn in
    let mn = if not !tofeatures then mn else
      let factor_to_fset f = Mn.Factor.FeatureSet (Mn.Factor.to_features f) in
      let factors = Array.map factor_to_fset (Mn.factors mn) in
      Mn.create mn.Mn.schema factors in
    Mn.output (open_out !outfile) mn in

  match (filetype !modelfile, filetype !outfile) with
  | (BNFile, BNFile) ->
    let bn = Bn.load_auto !modelfile in
    (* TODO: Convert trees to tables if necessary... *)
    let bn = if has_ev then Bn.simplify bn ev else bn in
    Bn.write_auto !outfile bn
  | (BNFile, MNFile) ->
    let bn = Bn.load_auto !modelfile in
    proc_mn (Bn.to_mn bn)
  | (MNFile, MNFile) ->
    proc_mn (Mn.load (open_in !modelfile))
  | (ACFile, MNFile) -> 
    let (ac, fl) = Circuit.load_with_features (open_in !modelfile) in
    if has_ev then Circuit.prune_for_evidence ac ev;
    let f_to_factor f =
      Mn.Factor.Feature
        {Mn.Factor.cond = f.Circuit.cond;
         Mn.Factor.weight_id = f.Circuit.acnode.Circuit.id;
         Mn.Factor.weight = log f.Circuit.weight} in
    let factors = Array.of_list (List.map f_to_factor fl) in
    let mn = Mn.create (ac.Circuit.schema) factors in
    proc_mn mn
  | (ACFile, ACFile) -> 
    let (ac, fl) = Circuit.load_with_features (open_in !modelfile) in
    if has_ev then Circuit.prune_for_evidence ac ev;
    let fl = if has_ev then Circuit.prune_features fl ev else fl in
    Circuit.output_with_features (open_out !outfile) ac fl
  | _ -> nlogf "ERROR: Unsupported conversion.\n"

;;
main ()
